%{
/*
   Dialogos - automate network communication
   Copyright (C) 2002 Markus Triska

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA

   Contact me at triska@gmx.at
*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include "node.h"
#include "dialogos.tab.h"
#include "network.h"

#define isatty(a) std::isatty(a)
extern int yyparse(void);
extern void freeProcedures();

int lineno = 1;
char buf[500];
char *s;
int strsize = 0;
%}
%x QSTRING
%%
APPEND       {   return APPEND;       }
CONNECT      {   return CONNECT;      }
CREATE       {   return CREATE;       }
DISCONNECT   {   return DISCONNECT;   }
ELSE         {   return ELSETOK;      }
FILE         {   return FILETOK;      }
FILESIZE     {   return FILESIZE;     }
GREP         {   return GREP;         }
HTTP_APPEND  {   return HTTP_APPEND;  }
HTTP_POST    {   return HTTP_POST;    }
HTTP_RECEIVE {   return HTTP_RECEIVE; }
IF           {   return IF;           }
IGNORE       {   return IGNORE;       }
LINEWISE     {   return LINEWISE;     }
NUMERIC      {   return NUMERIC;      }
PERFORM      {   return PERFORM;      }
PRINT        {   return PRINT;        }
PROCEDURE    {   return PROCEDURE;    }
PROCESS      {   return PROCESS;      }
RECEIVE      {   return RECEIVE;      }
SEND         {   return SEND;         }
SENDSTRING   {   return SENDSTRING;   }
SIZE         {   return SIZE;         }
SUBSTVARS    {   return SUBSTVARS;    }
USING        {   return USING;        }
WHILE        {   return WHILE;        }
">="         {   return GE;           }
"<="         {   return LE;           }
"=="         {   return EQ;           }
"NOT ="      {   return NE;           }

\"              { BEGIN QSTRING; s = buf;
                  strsize = 0;
                }
<QSTRING>\\n     { *s++ = '\n';  strsize++;   }
<QSTRING>\\r     { *s++ = '\r';  strsize++;   }
<QSTRING>\\t     { *s++ = '\t';  strsize++;   }
<QSTRING>\\\"    { *s++ = '\"';  strsize++;   }
<QSTRING>\"       {
			*s = 0;
			BEGIN 0;
			yylval.tok_string = (char *) malloc((strlen(buf) + 1) * sizeof(char));
			if (yylval.tok_string == NULL) {
				printf("malloc() of size %d failed\n", strlen(buf));
			} else {
				strcpy(yylval.tok_string, buf);
				return QUOTEDSTRING;
			}

                }
<QSTRING>.       { *s++ = *yytext; strsize++; }


[a-zA-Z_]+[0-9]* {
//		printf("Token: STRING\n");
		yylval.tok_string = (char *) malloc((strlen(yytext) + 1) * sizeof(char));
		if (yylval.tok_string == NULL) {
			printf("malloc() of size %d failed\n", strlen(yytext));
		} else {
			strcpy(yylval.tok_string, yytext);

			return STRING;
		}
	}

[0-9]* {
//		printf("Token: NUMBER\n");
		yylval.tok_ivalue = atoi(yytext);
		return NUMBER;
	}

[ \r\t]  {	//printf("Token: WHITESPACE\n");
	}

\n	{ lineno++; }

"!".*\r?\n {
//	  printf("Token: COMMENT\n");
	  lineno++; // skip comments
	}

.  {	return *yytext;	   }
%%

#include <map>
#include <string>
#include <iostream>

typedef std::map<std::string, std::string, std::less<std::string> > variables_type;

variables_type variables;

int yywrap()
{
	return 1;
}

int yyerror(char *s)
{
	printf("Error on line %d: %s\n", lineno, s);
	return 0;
}

using namespace std;

int main(int argc, char *argv[])
{
	cout << "Dialogos version 0.4, Copyright (C) 2002 Markus Triska" << endl;
	cout << "Dialogos comes with ABSOLUTELY NO WARRANTY. This is free software, " << endl;
	cout << "and you can redistribute it under the terms of the GNU GPL." << endl << endl;
	doStartup();
	if (argc >= 2) {
		yyin = fopen(argv[1], "r");
		if (!yyin) {
			cout << "Could not find file: " << argv[1] << endl;
			cout << "Reading from stdin instead." << endl;
		}
	} else {
		yyin = stdin;
	}

	variables["PORT"] = "80";
	variables["HOST"] = "localhost";
	variables["REFDELIM"] = "#";
	variables["VERBOSE"] = "FALSE";
	for (int i = 1; ((i < argc) && (i < 10)); i++) {
		char varname[5];
		sprintf(varname, "ARG%d", i);
		variables[varname] = argv[i];
	}

	yyparse();
	freeProcedures();
/*
	for (variables_type::iterator it = variables.begin(); it != variables.end(); it++) {
		pair<string, string> p = (*it);
		cout << "Variable: " << p.first << " is: " << p.second << endl;
	}  */
	doCleanup();
	return 0;
}



