%{
/*
   Dialogos - automate network communication
   Copyright (C) 2002 Markus Triska

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA

   Contact me at triska@gmx.at
*/


#include <malloc.h>
#include "node.h"

extern int yyerror(char *);
extern int yylex();
extern nodeType *opr(int, int , ...);
extern nodeType *str(char *);
extern nodeType *id(char *);
extern nodeType *con(int );
extern void registerProcedure(char *, nodeType *);
extern int ex(nodeType *);
extern void freeNode(nodeType *);
%}
%union {
	int tok_ivalue;
	char *tok_string;
	nodeType *nPtr;
}

%token <tok_ivalue> NUMBER
%token <tok_string> STRING QUOTEDSTRING
%token CONNECT HTTP_POST SEND RECEIVE HTTP_RECEIVE DISCONNECT ELSETOK FILESIZE
%token IGNORE LINEWISE PRINT NUMERIC PROCESS USING APPEND HTTP_APPEND IF GE
%token GREP LE EQ NE WHILE CREATE SIZE FILETOK SENDSTRING PROCEDURE PERFORM
%token SUBSTVARS

%left GE EQ NE LE '<' '>'
%left '+' '-'
%type <tok_string> variable
%type <nPtr> stmt stmt_list varassignment expr
%%

commands: /* empty */
       | commands PROCEDURE STRING '{' stmt_list '}' {
		registerProcedure($3, $5);                      }
       | commands stmt { ex($2); freeNode($2);                  }
	;

stmt:   APPEND QUOTEDSTRING QUOTEDSTRING {
		$$ = opr(APPEND, 2, str($2), str($3));          }
      | CONNECT    { $$ = opr(CONNECT, 0);                      }
      | CREATE QUOTEDSTRING { $$ = opr(CREATE, 1, str($2));     }
      | DISCONNECT { $$ = opr(DISCONNECT, 0);                   }
      | HTTP_APPEND QUOTEDSTRING {
		$$ = opr(HTTP_APPEND, 1, str($2));              }
      | HTTP_POST QUOTEDSTRING {
		$$ = opr(HTTP_POST, 1, str($2));                }
      | HTTP_RECEIVE QUOTEDSTRING {
		$$ = opr(HTTP_RECEIVE, 1, str($2));             }
      | HTTP_RECEIVE QUOTEDSTRING NUMBER {
		$$ = opr(HTTP_RECEIVE, 2, str($2), con($3));    }
      | HTTP_RECEIVE IGNORE {
		$$ = opr(HTTP_RECEIVE, 0);                      }
      | PRINT STRING {  $$ = opr(PRINT, 1, id($2));             }
      | PRINT QUOTEDSTRING { $$ = opr(PRINT, 1, str($2));       }
      | PRINT QUOTEDSTRING QUOTEDSTRING {
		$$ = opr(PRINT, 2, str($2), str($3));           }
      | RECEIVE QUOTEDSTRING { $$ = opr(RECEIVE, 1, str($2));   }
      | RECEIVE QUOTEDSTRING NUMBER {
		$$ = opr(RECEIVE, 2, str($2), con($3));         }
      | RECEIVE IGNORE { $$ = opr(RECEIVE, 0);                  }
      | SEND QUOTEDSTRING { $$ = opr(SEND, 1, str($2));         }
      | SENDSTRING QUOTEDSTRING {
		$$ = opr(SENDSTRING, 1, str($2));               }
      | PROCESS QUOTEDSTRING USING STRING '{' stmt_list '}' {
		$$ = opr(PROCESS, 3, str($2), id($4), $6);      }
      | PROCESS LINEWISE QUOTEDSTRING USING STRING '{' stmt_list '}' {
		$$ = opr(LINEWISE, 3, str($3), id($5), $7);     }
      | IF '(' expr ')' '{' stmt_list '}' {
		$$ = opr(IF, 2, $3, $6);                        }
      | IF '(' expr ')' '{' stmt_list '}' ELSETOK '{' stmt_list '}' {
		$$ = opr(ELSETOK, 3, $3, $6, $10);              }
      | WHILE '(' expr ')' '{' stmt_list '}' {
		$$ = opr(WHILE, 2, $3, $6);                     }
      | PERFORM STRING { $$ = opr(PERFORM, 1, str($2));         }
      | varassignment { $$ = $1;                                }
      ;

stmt_list: stmt            { $$ = $1;                           }
      | stmt_list stmt     { $$ = opr(';', 2, $1, $2);          }
      ;

varassignment:  variable '=' QUOTEDSTRING {
		$$ = opr('=', 2, id($1), str($3));                   }
      | variable '=' GREP QUOTEDSTRING variable {
		$$ = opr(GREP, 3, id($1), str($4), id($5));          }
      | variable '=' expr  { $$ = opr('=', 2, id($1), $3);           }
      | variable '=' SUBSTVARS '(' variable ')' {
		$$ = opr(SUBSTVARS, 2, id($1), id($5));              }
      | variable '=' SUBSTVARS '(' QUOTEDSTRING ')' {
		$$ = opr(SUBSTVARS, 2, id($1), str($5));             }
      | variable '=' FILETOK QUOTEDSTRING NUMBER {
		$$ = opr('=', 3, id($1), str($4), con($5));          }
      | variable '=' FILETOK QUOTEDSTRING NUMBER ':' NUMBER {
		$$ = opr('=', 4, id($1), str($4), con($5), con($7)); }
      | variable '=' FILETOK QUOTEDSTRING NUMBER '-' QUOTEDSTRING {
		$$ = opr('=', 4, id($1), str($4), con($5), str($7)); }
      | variable '=' FILETOK QUOTEDSTRING NUMBER NUMERIC {
		$$ = opr('=', 4, id($1), str($4), con($5), NULL);    }
      ;

expr:   NUMBER           { $$ = con($1);                      }
      | variable         { $$ = id ($1);                      }
      | SIZE '(' variable ')' {
 			$$ = opr(SIZE, 1, id($3));            }
      | FILESIZE '(' QUOTEDSTRING ')' {
			$$ = opr(FILESIZE, 1, str($3));       }
      | expr '+' expr    { $$ = opr('+', 2, $1, $3);          }
      | expr '-' expr    { $$ = opr('-', 2, $1, $3);          }
      | expr '<' expr    { $$ = opr('<', 2, $1, $3);          }
      | expr '>' expr    { $$ = opr('>', 2, $1, $3);          }
      | expr EQ expr     { $$ = opr(EQ, 2, $1, $3);           }
      | expr GE expr     { $$ = opr(GE, 2, $1, $3);           }
      | expr NE expr     { $$ = opr(NE, 2, $1, $3);           }
      | expr LE expr     { $$ = opr(LE, 2, $1, $3);           }

variable: STRING
      ;


