/*
   Dialogos - automate network communication
   Copyright (C) 2002 Markus Triska

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA

   Contact me at triska@gmx.at
*/


/* simplistic grep - not POSIX compliant */
/* April 10th, 2002 */

#include <string>
// #include <iostream>  // for debugging messages
#include <string.h>
#include <malloc.h>
#include <ctype.h>

static int NOMATCH = 0;
static int depth = -1;

using namespace std;


unsigned int nextcharindex(string s)
{
	unsigned int index = 0;
	char c;
	do {
		c = s[index++];
		if (c == '\\') break;
		if ((index - 1 ) >= s.size()) return s.size();
	} while ((c == '(') || (c == ')') || (c == '*') || (c == '?'));

	return --index;
}


int grep(string pattern, string line, string &result)
{
	unsigned int pi = 0, li = 0, lastlineoffset = 0;
	if (!pattern.size() || !line.size()) return NOMATCH;

	depth++;
	string depthstr = "";
	for (int i = 0; i < depth; i++) depthstr += "     ";

try_match:
	unsigned int fci = nextcharindex(pattern);
	if (fci >= pattern.size()) return NOMATCH;
	char c = pattern[fci];

	string refmatch = "";

	if (c == '\\') {
		if ((fci + 1) >= pattern.size()) return NOMATCH;
		c = pattern[fci + 1];
	}

	bool insideref = false;
	int nopenbraces = 0;
	for (unsigned int i = 0; i < fci; i++) {
		if (pattern[i] == '(') nopenbraces++;
		if (pattern[i] == ')') nopenbraces--;
	}
	if (nopenbraces > 0) insideref = true;


	if (depth == 0) {
		// scan to next occurance of the first character and try the pattern
		c = toupper(c);
		do {
			if (toupper(line[li]) == c) break;
			// This is for patterns like "(*)anything":
			// (on the other hand, it will also add everything up to
			// "anything" when the pattern is "()anything")
			if (pattern[0] == '(') refmatch += line[li];
			li++;
		} while (li < line.size());
	}

	if (li >= line.size()) return NOMATCH;

	lastlineoffset = li;
//	cout << depthstr << "Pattern: " << pattern << endl;
//	cout << depthstr << "From beginmatch: " << li << " " << line.substr(li) << endl;

	while (pi < pattern.size()) {
		c = pattern[pi++];
//		cout << depthstr << "Comparing " << c << " to: " << line[li] << endl;
		switch (c) {
		case '(': insideref = true;
			li--;
			break;
		case ')': insideref = false;
			li--;
			break;

		case '\\':
			if (line[li] != pattern[pi++]) goto nomatch;
			if (insideref) refmatch += line[li];
			break;

		case '*':
			if (pi >= pattern.size()) {
				if (insideref) {
					while (li < line.size()) refmatch += line[li++];
				}
				goto matched;
			}
			{
				bool dogreedy = true;
				if (pattern[pi] == '?') {
//					cout << "Doing nongreedy search" << endl;
					dogreedy = false;
					pi++;
				}

				// Find next character to match in pattern
				string partpattern = pattern.substr(pi);
				unsigned int fci1 = nextcharindex(partpattern);

				if (fci1 >= partpattern.size()) {
					// wildcard was last - rest is match
					while (li < line.size()) {
						if (insideref) refmatch += line[li];
						li++;
					}
					goto matched;
				}

				fci1 += pi;
				char c1 = pattern[fci1];
				if (c1 == '\\') {
					if ((fci1 + 1) >= pattern.size()) return NOMATCH;
					c1 = pattern[fci1 + 1];
				}



//				cout << depthstr << "Next character after wc: " << c1 << " (" << toupper(c1) << ")" << endl;
				c1 = toupper(c1);
				// Begin and end of part of line that
				// matched the wildcard
				unsigned int begin_wc = li;
				unsigned int end_wc;
				unsigned int bli = li;
				string wcmatch = "";
				bool matchfound = false;

				// Scan to position where line[pos] == c1 and
				// try to apply the rest of the pattern. This
				// will fail to work as expected if the very
				// next character after the wildcard is
				// optional!
				// Such a specification, however, is useless and
				// therefore (I hope) rare.

				while (bli < line.size()) {
					if (toupper(line[bli]) == c1) {
//						cout << depthstr << "Trying greedy match to "<< c1 << endl;
//						cout << depthstr << "Recursion for pattern: " << pattern.substr(pi) << " in: " << line.substr(bli) << endl;
						string localresult;
						if (grep(pattern.substr(pi), line.substr(bli), localresult)) {
							end_wc = bli;
							matchfound = true;
							wcmatch = string(line, begin_wc, end_wc - begin_wc);
//							cout << depthstr << "Applied pattern, found: " << wcmatch << endl;
							li = bli - 1;
							if (!dogreedy) goto no_greedy;
						} else {
//							cout << depthstr << "Could not apply pattern." << endl;
						}
					}
					bli++;
				}
				if (!matchfound) goto nomatch;
				no_greedy:
				if (insideref) refmatch += wcmatch;
				if (pattern[pi] == ')') insideref = false;
			}
			break;

		default:
			if (pi < pattern.size()) {
				if (pattern[pi] == '?') {
//					cout << "Optional: " << c << endl;
					// optional match
					if (toupper(c) == toupper(line[li])) {
						if (insideref) refmatch += line[li];
					} else {
						// back up to last position
						li--;
					}

					pi++;
					break;
				}
			}
			if (toupper(c) != toupper(line[li])) goto nomatch;
			if (insideref) refmatch += line[li];
			break;
		}
		li++;
	}
	goto matched;

nomatch:
//	cout << depthstr << "Reached 'nomatch'" << endl;
	li = lastlineoffset + 1;
	pi = 0;
	if (li < line.size() && depth == 0) goto try_match;
	depth--;
	return NOMATCH;

matched:
//	cout << depthstr << "Matched: " << refmatch << endl;
	result += refmatch;
	depth--;
	return 1;
}

