/*
   Dialogos - automate network communication
   Copyright (C) 2002 Markus Triska

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA

   Contact me at triska@gmx.at
*/


/* Dialogos networking functions */

#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <stdio.h>

#include "network.h"

#include <string>
#include <iostream>
#include <sstream>
#include <fstream>
#include <algorithm>
#ifdef WINDOWS
#include <windows.h>
#include <winsock.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <netdb.h>
#endif

using namespace std;

extern string substVars(string );

static int servfd;
static char *servip;
static bool connected = false;

int doStartup()
{
#ifdef WINDOWS
	WSADATA wsadata;
	WSAStartup(MAKEWORD(1, 1), &wsadata);
#endif
	return 0;
}

bool doVerbose()
{
	return (variables["VERBOSE"] == "TRUE");
}



int doConnect()
{
	variables_type::iterator it = variables.find("IP");
	if (it != variables.end()) {
		servip = (char *) (*it).second.c_str();
		variables["SERVIP"] = servip;
	} else {
		string hostname = "Nohost";
		variables_type::iterator it = variables.find("HOST");
		if (it != variables.end()) {
			hostname = substVars((*it).second);
		}

		if (doVerbose()) cout << "Looking up host '" << hostname << "'" << endl;
		struct hostent *hptr = gethostbyname(hostname.c_str());
		if (hptr == NULL) {
			cout << "Could not find host '" << hostname << "'" << endl;
			return -1;
		}

		in_addr addr;
		memcpy(&addr, hptr->h_addr_list[0], sizeof(struct in_addr));
		servip = inet_ntoa(addr);
		variables["SERVIP"] = servip;
	}

	short port = 80;
	variables_type::iterator pt = variables.find("PORT");
	if (pt != variables.end()) {
		 port = (short) atoi((*pt).second.c_str());
	}

	if (doVerbose()) cout << "Initializing connection to IP " << servip << ", port " << port << "..." << endl;

	sockaddr_in servaddr;
	servfd = socket(AF_INET, SOCK_STREAM, 0);
	if (servfd == INVALID_SOCKET) {
		cout << "Could not create socket." << endl;
		return -1;
	}
	memset(&servaddr, 0, sizeof(servaddr));
	servaddr.sin_family = AF_INET;
	servaddr.sin_port = htons(port);
	servaddr.sin_addr.s_addr = inet_addr(servip);

	if (connect(servfd, (struct sockaddr *) &servaddr, sizeof(servaddr)) < 0) {
		cout << "Connect() failed." << endl;
		return -1;
	}
	connected = true;
	return 1;
}


string inputHeader(ifstream &infile)
{
	string header;
	string oneline;

	while (!infile.eof()) {
		char c = infile.get();
		if (!infile.eof()) {
			if (c == '\n') {
				oneline += "\15\12";
				header += oneline;
				if (oneline.size() == 2) {
					return header;
				}
				oneline = "";
			} else {
				oneline += c;
			}
		}
	}
	header += "\15\12";
	return header;
}

int httpPost(string file)
{
	if (!connected) {
		cout << "HTTP_POST: Not connected to any server." << endl;
		return -1;
	}

	ifstream infile(file.c_str());
	if (!infile) {
		cout << "HTTP_POST: Cannot find file " << file << endl;
		return -1;
	} else {
		if (doVerbose()) cout << "Sending contents of file " <<  file << endl;
	}

	string header = inputHeader(infile);
	string body = "";
	while (!infile.eof()) {
		char c = infile.get();
		if (!infile.eof()) {
			body += c;
		}
	}

	char buf[20];
	body = substVars(body);
	int cutit = 0;
	if (body.size() > 0) {
		if (body[body.size() - 1] == '\n') {
			cutit = 1;
		}
	}
	sprintf(buf, "%d", body.size() - cutit);
	variables["CONTENTLENGTH"] = buf;

	string message = substVars(header) + string(body.begin(), body.end() - cutit);
//	cout << "Sending: " << message << endl;
	send(servfd, message.c_str(), message.size(), 0);
	return 1;
}

int receiveByte(char *c, bool doreceive = true, int maxbytes = BUFFERSIZE)
{
	static char buffer[BUFFERSIZE];
	static int readpos, writepos;

	if (readpos == writepos) {
		readpos = 0;
		if (doreceive) {
			writepos = recv(servfd, buffer, maxbytes, 0);
//			cout << "receiveByte() has received " << writepos << "bytes" << endl;
			*c = buffer[readpos++];
		} else {
			writepos = 0;
		}
		return writepos;
	}

	*c = buffer[readpos++];
	return writepos;
}

int receiveLine(string &s, int maxbytes = BUFFERSIZE)
{
	char c;
	bool crseen = false, lfseen = false;
	while (!lfseen) {
		receiveByte(&c, true, maxbytes);
		if (c == '\15') {
			crseen = true;
		} else if (c == '\12') {
			if (crseen) lfseen = true;
		} else {
			crseen = false;
			s += c;
		}
	}
	return 0;
}

int httpSkipHeader()
{
	int content_length = 0;
	int maxbytes = BUFFERSIZE;
	variables_type::iterator it = variables.find("HEADER_LENGTH");

	// Make sure not to reach end of HTTP-reply when searching for
	// "Content-Length". This is critical because the next recv would
	// otherwise block forever.
	if (it != variables.end()) {
		int header_length = atoi(variables["HEADER_LENGTH"].c_str());
		if (header_length < maxbytes) {
			maxbytes = header_length;
		}
	}

	while (!content_length) {
		string oneline;
		receiveLine(oneline, maxbytes);
//		if (oneline.size()) cout << "Inspecting " << oneline << " for content-length" << endl;
		if (oneline.size() == 0) {
			// reached end of header -- HTTP/1.0 reply
			return -1;
		}
		if (string(oneline.begin(), oneline.begin() + 15) == string("Content-Length:")) {
			istringstream clstring (oneline.substr(16));
			clstring >> content_length;
		}
	}
	// Skip rest of header
	string oneline;
	do {
		oneline = "";
		receiveLine(oneline);
//		cout << "Skipping one line: " << oneline << endl;
	} while (oneline.size());
	return content_length;
}

int httpReceive(string file, int numsavebytes, bool append)
{
	if (!connected) {
		cout << "HTTP_RECEIVE: not connected to any server" << endl;
		return -1;
	}

	replace(file.begin(), file.end(), '/', '_');
	int content_length = httpSkipHeader();

	// The rest of the message is the actual reply
	int numbytesread = 0, numbyteswritten = 0;
	char c;
	string reply = "";

	// read in whatever might still be in the buffer
	while (receiveByte(&c, false)) {
		reply += c;
		numbytesread++;
		numbyteswritten++;
	}
//	cout << numbytesread << " were still in the buffer: " << reply.size() << endl ;

	ofstream *replyfile = NULL;
	if (file.size()) {
		if (doVerbose()) cout << "Saving body of HTTP-reply to " << file << endl;
		if (append) {
			replyfile = new ofstream(file.c_str(), ios::app | ios::binary);
		} else {
			replyfile = new ofstream(file.c_str(), ios::binary);
		}

		*replyfile << reply;
	} else {
		if (doVerbose()) cout << "Ignoring reply " << endl;
	}

	// Now read until we reach content_length bytes or peer closes socket
	for (; ;) {
		int maxread = BUFFERSIZE;
		if (content_length > -1) {
			maxread = content_length - numbytesread;
			if (maxread > BUFFERSIZE) {
				maxread = BUFFERSIZE;
			}
		}
		char buffer[BUFFERSIZE];
		int numread = recv(servfd, buffer, maxread, 0);
		numbytesread += numread;
//		cout << "Read: 	" << numread << " " << string(buffer, numread).size() << endl;
		if (numread <= 0) {
			closesocket(servfd);
			servfd = INVALID_SOCKET;
			connected = false;
			if (numread < 0) {
				cout << "Connection died" << endl;
			}
			break;
		}

		if (replyfile) {
			bool dowrite = true;
			if (numsavebytes && (numbyteswritten >= numsavebytes)) {
				dowrite = false;
			}
			if (dowrite) {
				*replyfile << string(buffer, numread);
				numbyteswritten = numbytesread;
			}
		}

		if (content_length > -1) {
			if (numbytesread >= content_length) break;
		}

	}
//	cout << "Read " << numbytesread << " bytes in total." << endl;
	if (replyfile) {
		replyfile->close();
		delete replyfile;
	}
	return 1;
}

int doReceive(string file, int numsavebytes, bool append)
{
	if (!connected) {
		cout << "RECEIVE: not connected to any server" << endl;
		return -1;
	}

	ofstream *replyfile = NULL;
	int numbytesread = 0, numbyteswritten = 0;

	replace(file.begin(), file.end(), '/', '_');
	if (file.size()) {
		if (doVerbose()) cout << "Saving reply to " << file << endl;
		if (append) {
			replyfile = new ofstream(file.c_str(), ios::app | ios::binary);
		} else {
			replyfile = new ofstream(file.c_str(), ios::binary);
		}
	} else {
		if (doVerbose()) cout << "Ignoring reply " << endl;
	}

	int numread;
	char buffer[BUFFERSIZE];
	while ( (numread = recv(servfd, buffer, BUFFERSIZE, 0)) > 0) {
		if (numread <= 0) {
			closesocket(servfd);
			servfd = INVALID_SOCKET;
			connected = false;
			if (numread < 0) {
				cout << "Connection died" << endl;
			}
			break;
		}

		numbytesread += numread;
		if (replyfile) {
			bool dowrite = true;
			if ((numsavebytes > 0) && (numbyteswritten > numsavebytes)) {
				dowrite = false;
			}
			if (dowrite) {
				*replyfile << string(buffer, numread);
				numbyteswritten = numbytesread;
			}
		}
	}
	closesocket(servfd);
	connected = false;
	servfd = INVALID_SOCKET;

	if (replyfile) {
		replyfile->close();
		delete replyfile;
	}
	return 1;
}

int sendString(string s)
{
	if (!connected) {
		cout << "SENDSTRING: not connected to any server." << endl;
		return -1;
	}

	string tosend = substVars(s);
	send(servfd, tosend.c_str(), tosend.size(), 0);
	return 1;
}

int sendFile(string file)
{
	if (!connected) {
		cout << "SEND: not connected to any server." << endl;
		return -1;
	}
	ifstream infile(file.c_str());
	if (!infile) {
		cout << "SEND: file " << file << " not found." << endl;
		return -1;
	}

	if (doVerbose()) cout << "Sending contents of " << file << endl;
	string tosend;
	while (!infile.eof()) {
		char c = infile.get();
		if (!infile.eof()) {
			tosend += c;
		}
	}
	tosend = substVars(tosend);
	if (send(servfd, tosend.c_str(), tosend.size(), 0) < 0) {
		cout << "SEND failed!" << endl;
	}
	infile.close();
	return 1;
}

int closeConnection()
{
	if (servfd != INVALID_SOCKET) {
		if (doVerbose()) cout << "Closing connection." << endl;
		shutdown(servfd, 1);
		char buffer[200];
		while (recv(servfd, buffer, 200, 0) > 0) {
			// Wait until peer shuts down
		}
		closesocket(servfd);
		servfd = INVALID_SOCKET;
	}
	connected = false;
	return 0;
}

int doCleanup()
{
	if (connected) closeConnection();
#ifdef WINDOWS
	WSACleanup();
#endif
	return 1;
}
